/*
  TXboxAdpcmDecoder class
  (c) 2005 Benjamin Haisch

  Revision 2 with stereo support

###
  C conversion 0.1.3
  by Luigi Auriemma
  e-mail: aluigi@autistici.org
  web:    aluigi.org

    Copyright 2005,2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
###
  C# conversion
  by FastElbJa
  e-mail: fastelbja@tosec.org
  
*/


#include <windows.h>
#include "wamain.h"
#include "cube.h"

#define XBOX_ADPCM_SRCSIZE 36
#define XBOX_ADPCM_DSTSIZE 130

const static s16 StepTable[89] = {  7,     8,     9,    10,    11,    12,    13,    14,    16,    17,
                        19,    21,    23,    25,    28,    31,    34,    37,    41,    45,
                        50,    55,    60,    66,    73,    80,    88,    97,   107,   118,
                       130,   143,   157,   173,   190,   209,   230,   253,   279,   307,
                       337,   371,   408,   449,   494,   544,   598,   658,   724,   796,
                       876,   963,  1060,  1166,  1282,  1411,  1552,  1707,  1878,  2066,
                      2272,  2499,  2749,  3024,  3327,  3660,  4026,  4428,  4871,  5358,
                      5894,  6484,  7132,  7845,  8630,  9493, 10442, 11487, 12635, 13899,
                     15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767};
    
const static s8 IndexTable[16] = { -1, -1, -1, -1, 2, 4, 6, 8,
						-1, -1, -1, -1, 2, 4, 6, 8};

int TXboxAdpcmDecoder_DecodeSample(int Code, CUBESTREAM *CurrentChannel) 
{
    s32 Delta;
    s32 Result;

    Delta = (CurrentChannel->StepSize >> 3);
    if ((Code & 4)==4) Delta += CurrentChannel->StepSize;
    if ((Code & 2)==2) Delta += CurrentChannel->StepSize >> 1;
    if ((Code & 1)==1) Delta += CurrentChannel->StepSize >> 2;
    if ((Code & 8)==8) Delta = -Delta;
    Result = (CurrentChannel->Predictor + Delta);
	if (Result==0xffff8bee)
		Delta=0;

	//if ((unsigned int)Result==0x8275)
	//	Delta =0;

    if (Result >= 32767)
    {
        Result =  32767;
    }
    else if (Result <= -32768)
    {
        Result = -32768;
    }

    CurrentChannel->index += IndexTable[Code];

    if (CurrentChannel->index > 88)
    {
        CurrentChannel->index = 88;
    }
    else if (CurrentChannel->index < 0)
    {
        CurrentChannel->index = 0;
    }

    CurrentChannel->StepSize  = StepTable[CurrentChannel->index];
    CurrentChannel->Predictor = Result;
    return(Result);
}
    
void fillBufferXBOX(CUBEFILE *xbox)
{
    unsigned char buffer[0x800];
	int	bytesRead;

    unsigned int NbSamplesPlayed = 0;
	int CodeBuf=0;
	int i,j;
    int bufferOffset = 0;

	//xbox->ch[0].writeloc=xbox->ch[1].writeloc=0;

	if ((SetFilePointer(xbox->ch[0].infile,0,0,FILE_CURRENT) - xbox->ch[0].chanstart) > xbox->file_length)
		return;

	bufferOffset=0;

	SetFilePointer(xbox->ch[0].infile,xbox->ch[0].offs,0,FILE_BEGIN);
	ReadFile(xbox->ch[0].infile, buffer,0x48,&bytesRead,NULL);
	
	xbox->ch[0].offs += (36 * xbox->NCH);

	xbox->ch[0].Predictor = (s16)(buffer[bufferOffset] | (buffer[bufferOffset+1] << 8)) ;
	xbox->ch[0].index = (s16)(buffer[bufferOffset + 2] | (buffer[bufferOffset + 3] << 8));
	xbox->ch[0].chanbuf[xbox->ch[0].writeloc++] = (short)xbox->ch[0].Predictor;
	if (xbox->ch[0].index > 88)
		xbox->ch[0].index = 88;
	else if (xbox->ch[0].index < 0)
		xbox->ch[0].index = 0;
	xbox->ch[0].StepSize = StepTable[xbox->ch[0].index];
	bufferOffset += 4;

	if (xbox->NCH == 2)
	{
		xbox->ch[1].Predictor = (s16)(buffer[bufferOffset] | (buffer[bufferOffset+1] << 8)) ;
		xbox->ch[1].index = (s16)(buffer[bufferOffset + 2] | (buffer[bufferOffset + 3] << 8));
		xbox->ch[1].chanbuf[xbox->ch[1].writeloc++] = (short)xbox->ch[1].Predictor;
		if (xbox->ch[1].index > 88)
			xbox->ch[1].index = 88;
		else if (xbox->ch[1].index < 0)
			xbox->ch[1].index = 0;
		xbox->ch[1].StepSize = StepTable[xbox->ch[1].index];
		bufferOffset += 4;
	}
    

	for(i = 0; i < 8; i++) 
	{
		CodeBuf = buffer[bufferOffset] |
				 (buffer[bufferOffset + 1] << 8) |
				 (buffer[bufferOffset + 2] << 16) |
				 (buffer[bufferOffset + 3] << 24);
		for(j = 0; j < 8; j++) 
		{
			xbox->ch[0].chanbuf[xbox->ch[0].writeloc] = (short)
				TXboxAdpcmDecoder_DecodeSample(CodeBuf & 0x0F, &xbox->ch[0]);
			
			xbox->ch[0].writeloc++;
			if (xbox->ch[0].writeloc>=BUFFER_SIZE/8*14) xbox->ch[0].writeloc=0;

			CodeBuf >>= 4;
		}

		if (xbox->NCH == 2)
		{
			bufferOffset += 4;
			CodeBuf = buffer[bufferOffset] |
					 (buffer[bufferOffset + 1] << 8) |
					 (buffer[bufferOffset + 2] << 16) |
					 (buffer[bufferOffset + 3] << 24);
			for(j = 0; j < 8; j++) 
			{
				xbox->ch[1].chanbuf[xbox->ch[1].writeloc] = 
					TXboxAdpcmDecoder_DecodeSample(CodeBuf & 0x0F, &xbox->ch[1]);
				xbox->ch[1].writeloc++;
				if (xbox->ch[1].writeloc>=BUFFER_SIZE/8*14) xbox->ch[1].writeloc=0;
				CodeBuf >>= 4;
			}
		}
		bufferOffset += 4;
	}
	
}

int InitXBOX(char * inputfile, CUBEFILE * xbox)
{
	unsigned char readbuf[0x80];
	char tampon[4];
	int l;

	if (inputfile) 
	{
		xbox->ch[0].infile=xbox->ch[1].infile=INVALID_HANDLE_VALUE;
		xbox->ch[0].readloc=xbox->ch[0].writeloc=0;
		xbox->ch[1].readloc=xbox->ch[1].writeloc=0;
		xbox->ch[0].ea=xbox->ch[1].ea=0;
		xbox->ch[0].loop_flag=xbox->ch[1].loop_flag=0;
		xbox->ch[0].chanstart=0;

		if (!strcmpi(inputfile+strlen(inputfile)-5,".wavm")) 
		{

			xbox->ch[0].infile = CreateFile(inputfile,GENERIC_READ,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,
				OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);

			if (xbox->ch[0].infile == INVALID_HANDLE_VALUE) // error opening file
				return 1;

			xbox->ch[0].Predictor=xbox->ch[1].Predictor=0;
			xbox->ch[0].StepSize=xbox->ch[1].StepSize=0;
			xbox->ch[0].index=xbox->ch[1].index=0;
			xbox->ch[0].Predictor=xbox->ch[1].Predictor=0;
			xbox->ch[0].type=type_xbox;
			xbox->NCH=1;
			xbox->ch[0].sample_rate=xbox->ch[1].sample_rate=44100;
			xbox->file_length=GetFileSize(xbox->ch[0].infile,NULL);
			xbox->nrsamples= ((xbox->file_length / (36 * xbox->NCH)) * 130)/2;
			xbox->ch[0].offs=0;
			xbox->ch[0].chanstart=0;
			return 0;
		}
		else
		{
			xbox->ch[0].infile = CreateFile(inputfile,GENERIC_READ,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,
				OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);

			if (xbox->ch[0].infile == INVALID_HANDLE_VALUE) // error opening file
				return 1;
	
			ReadFile(xbox->ch[0].infile,readbuf,0x80, &l,NULL);
			xbox->file_length=GetFileSize(xbox->ch[0].infile,NULL);

			if (!memcmp(readbuf,"xbaud",5))
			{
				xbox->ch[0].type=type_xmu;
                xbox->NCH = get16bitL(readbuf+0x0a);
				xbox->ch[0].sample_rate = get32bitL(readbuf + 0x0c);
                xbox->nrsamples = ((xbox->file_length / 36 ) * 130)/xbox->NCH;
                xbox->ch[0].offs = xbox->ch[0].chanstart= 0x20;
				return 0;
			} else if(!memcmp(readbuf,"FSB",3)) {
				// FSB XBOX
				if(get32bitL(readbuf+0x48)== (int)0x00480041)
				{
					xbox->ch[0].type = type_xbfsb;
					xbox->NCH = 2;
					xbox->file_length = get32bitL(readbuf + 0x0c) + 0x58;
					xbox->nrsamples = (xbox->file_length / 36 * 130)/4;
					xbox->ch[0].sample_rate=xbox->ch[1].sample_rate=get32bitL(readbuf+0x4c);
                    if(get32bitL(readbuf+0x14)==0x0030001)
                        xbox->ch[0].offs = 0x68;
                    else
                        xbox->ch[0].offs = 0x58;
                    return 0;
				}
			}
			else if (!memcmp(readbuf,"RIFF",4))
			{
				if (!memcmp(readbuf+0x08,"WAVEfmt",7))
				{
					if (get16bitL(readbuf+0x14) == (int)0x0069)
					{
						xbox->ch[0].type=type_xbox;
						xbox->NCH = get16bitL(readbuf+0x16);
						xbox->ch[0].sample_rate=get32bitL(readbuf+0x18);

						// Search for "data"
						SetFilePointer(xbox->ch[0].infile,0x1C,0,FILE_BEGIN);
						do
						{
							ReadFile(xbox->ch[0].infile,tampon,4,&l,NULL);
							if(!memcmp(tampon,"data",4))
								break;
						} while (l!=0);
						
						if (l==0)
						{
							CloseCUBEFILE(xbox);
							return 1;
						}

						xbox->ch[0].offs=xbox->ch[0].chanstart=SetFilePointer(xbox->ch[0].infile,0,0,FILE_CURRENT)+4;
						ReadFile(xbox->ch[0].infile,tampon,4,&l,NULL);
						xbox->nrsamples=(xbox->file_length / 36 * 130)/4;
						return 0;

					}
				}
			}
		}
	}
	
	CloseCUBEFILE(xbox);
	return 1;
}